import json
from pathlib import Path
from typing import Any, Dict

import pandas as pd

from sktmls.autogluon import TabularPrediction as task, __version__

MLS_MODEL_DIR = Path.home().joinpath("models")


class MLSTrainable:
    def fit(
        self,
        train_data: pd.DataFrame,
        label: str,
        test_data: pd.DataFrame = None,
        gbm_hyperparameters: Dict[str, Any] = dict(),
        cat_hyperparameters: Dict[str, Any] = dict(),
        eval_metric: str = "roc_auc",
    ) -> None:

        self.model_lib = "autogluon"
        self.model_lib_version = __version__
        self.metric = eval_metric

        model_binary = task.fit(
            train_data=train_data,
            label=label,
            presets=["good_quality_faster_inference_only_refit"],
            hyperparameters={"GBM": gbm_hyperparameters, "CAT": cat_hyperparameters},
            eval_metric=eval_metric,
            auto_stack=True,
            output_directory=MLS_MODEL_DIR.joinpath(self.model_name, self.model_version),
        )

        model_binary.delete_models(models_to_keep="best", dry_run=False)
        model_binary.save_space(remove_data=True, remove_fit_stack=False, requires_save=False, reduce_children=False)

        self.models[0] = model_binary

        if test_data is not None:
            self.performance = self.evaluate(test_data)
            self.feature_importance = json.dumps(self.get_feature_importance(test_data).to_dict())

    def evaluate(self, test_data: pd.DataFrame) -> float:
        return self.models[0].evaluate(test_data, silent=True).tolist()

    def get_feature_importance(self, test_data: pd.DataFrame) -> pd.Series:
        return self.models[0].feature_importance(test_data, silent=True)

    def set_mms_path(self) -> None:
        trainer = self.models[0]._learner.load_trainer()
        for model_name in trainer.get_model_names_all():
            trainer.set_model_attribute(
                model_name, "path", f"/models/{self.model_name}/{self.model_version}/models/{model_name}/"
            )

    def set_local_path(self) -> None:
        trainer = self.models[0]._learner.load_trainer()
        for model_name in trainer.get_model_names_all():
            trainer.set_model_attribute(
                model_name, "path", f"{MLS_MODEL_DIR}/{self.model_name}/{self.model_version}/models/{model_name}/"
            )



import math
import numpy as np

from .catboost_utils import CustomMetric
from ....metrics import soft_log_loss
from .....try_import import try_import_catboost, try_import_catboostdev


# Ojectives for SOFTCLASS problem_type
# TODO: these require catboost_dev for now. Swap catboost_dev for catboost>=0.24 once it is released. Lazy imports will no longer be needed at that time.
class SoftclassCustomMetric(CustomMetric):
    try_import_catboost()  # Need to first import catboost then catboost_dev not vice-versa.
    try_import_catboostdev()
    from catboost_dev import MultiRegressionCustomMetric
    def __init__(self, metric, is_higher_better, needs_pred_proba):  # metric is ignored
        super().__init__(metric, is_higher_better, needs_pred_proba)
        try_import_catboostdev()
        self.softlogloss = self.SoftLogLossMetric()  # the metric object to pass to CatBoostRegressor

    def evaluate(self, approxes, target, weight):
        return self.softlogloss.evaluate(approxes, target, weight)

    class SoftLogLossMetric(MultiRegressionCustomMetric):
        def get_final_error(self, error, weight):
            return error

        def is_max_optimal(self):
            return True

        def evaluate(self, approxes, target, weight):
            assert len(target) == len(approxes)
            assert len(target[0]) == len(approxes[0])
            weight_sum = len(target)
            approxes = np.array(approxes)
            approxes = np.exp(approxes)
            approxes = np.multiply(approxes, 1/np.sum(approxes, axis=1)[:, np.newaxis])
            error_sum = soft_log_loss(np.array(target), np.array(approxes))
            return error_sum, weight_sum


class SoftclassObjective(object):
    try_import_catboost()  # Need to first import catboost then catboost_dev not vice-versa.
    try_import_catboostdev()
    from catboost_dev import MultiRegressionCustomObjective
    def __init__(self):
        try_import_catboostdev()
        self.softlogloss = self.SoftLogLossObjective()  # the objective object to pass to CatBoostRegressor

    class SoftLogLossObjective(MultiRegressionCustomObjective):
        # TODO: Consider replacing with C++ implementation (but requires building catboost from source).
        # This pure Python is 3x faster than optimized Numpy implementation. Tested C++ implementation was 3x faster than this one.
        def calc_ders_multi(self, approxes, targets, weight):
            exp_approx = [math.exp(val) for val in approxes]
            exp_sum = sum(exp_approx)
            exp_approx = [val / exp_sum for val in exp_approx]
            grad = [(targets[j] - exp_approx[j])*weight for j in range(len(targets))]
            hess = [[(exp_approx[j] * exp_approx[j2] - (j==j2)*exp_approx[j]) * weight
                    for j in range(len(targets))] for j2 in range(len(targets))]
            return (grad, hess)




import re

import numpy as np
import pandas as pd
from pandas import DataFrame


def make_date(df: DataFrame, date_field: str):
    """Make sure `df[field_name]` is of the right date type."""
    field_dtype = df[date_field].dtype
    if isinstance(field_dtype, pd.core.dtypes.dtypes.DatetimeTZDtype):
        field_dtype = np.datetime64
    if not np.issubdtype(field_dtype, np.datetime64):
        df[date_field] = pd.to_datetime(df[date_field], infer_datetime_format=True)


def add_datepart(df: DataFrame, field_name: str, prefix: str = None, drop: bool = True, time: bool = False):
    """Helper function that adds columns relevant to a date in the column `field_name` of `df`."""
    make_date(df, field_name)
    field = df[field_name]
    prefix = ifnone(prefix, re.sub('[Dd]ate$', '', field_name))
    attr = [
        'Year', 'Month', 'Week', 'Day', 'Dayofweek', 'Dayofyear', 'Is_month_end', 'Is_month_start',
        'Is_quarter_end', 'Is_quarter_start', 'Is_year_end', 'Is_year_start'
    ]
    if time:
        attr = attr + ['Hour', 'Minute', 'Second']
    for n in attr:
        df[prefix + n] = getattr(field.dt, n.lower())
    df[prefix + 'Elapsed'] = field.astype(np.int64) // 10 ** 9
    if drop:
        df.drop(field_name, axis=1, inplace=True)
    return df


def ifnone(a, b):
    """`a` if `a` is not None, otherwise `b`."""
    return b if a is None else a



from typing import Any, Dict, List

from sktmls.models import MLSLightGBMModel, MLSXGBoostModel, MLSCatBoostModel, MLSGenericModel


class DefaultLightGBMModel(MLSLightGBMModel):
    """
    MLS 모델 레지스트리에 등록되는 LightGBM 기본 클래스입니다.

    온라인 추론 없이 MLS 모델 레지스트리에 바이너리만을 등록하고자 할 때 사용합니다.
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        """
        본 클래스의 predict 함수는 아무런 일을 하지 않습니다.
        """
        return {"items": []}


class DefaultXGBoostModel(MLSXGBoostModel):
    """
    MLS 모델 레지스트리에 등록되는 XGBoost 기본 클래스입니다.

    온라인 추론 없이 MLS 모델 레지스트리에 바이너리만을 등록하고자 할 때 사용합니다.
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        """
        본 클래스의 predict 함수는 아무런 일을 하지 않습니다.
        """
        return {"items": []}


class DefaultCatBoostModel(MLSCatBoostModel):
    """
    MLS 모델 레지스트리에 등록되는 CatBoost 기본 클래스입니다.

    온라인 추론 없이 MLS 모델 레지스트리에 바이너리만을 등록하고자 할 때 사용합니다.
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        """
        본 클래스의 predict 함수는 아무런 일을 하지 않습니다.
        """
        return {"items": []}


class DefaultGenericModel(MLSGenericModel):
    """
    MLS 모델 레지스트리에 등록되는 일반모델 기본 클래스입니다.

    온라인 추론 없이 MLS 모델 레지스트리에 바이너리만을 등록하고자 할 때 사용합니다.
    models는 list 타입으로 모델이 한 개만 저장되는 경우에도 list 형태로 전달하여야 합니다.
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        """
        본 클래스의 predict 함수는 아무런 일을 하지 않습니다.
        """
        return {"items": []}



import setuptools
import os

SKTMLS_VERSION = os.getenv("SKTMLS_VERSION")
if not SKTMLS_VERSION:
    raise TypeError("NO SKTMLS_VERSION")


def load_long_description():
    with open("README.md", "r") as f:
        long_description = f.read()
    return long_description


setuptools.setup(
    name="sktmls",
    version=SKTMLS_VERSION,
    author="SKTMLS",
    author_email="mls@sktai.io",
    description="MLS SDK",
    long_description=load_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/sktaiflow/mls-sdk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "numpy>=1.19.4,<1.20",
        "catboost>=0.24.3,<0.25.0",
        "joblib",
        "lightgbm>=2.3.1,<2.4.0",
        "pandas>=1.1.1,<1.2.0",
        "pytz",
        "requests",
        "scikit-learn>=0.23.2,<0.24",
        "scipy>=1.4.1,<1.5.0",
        "xgboost>=1.2.1,<1.3",
        "python-dateutil",
        "matplotlib>=3.2.1,<3.3",
        "tqdm>=4.51.0",
        "boto3",
        "ConfigSpace<=0.4.10",
        "cryptography>=3.2.1",
        "cython",
        "dask>=2.6.0",
        "distributed>=2.6.0",
        "fastparquet==0.4.1",
        "gluoncv>=0.5.0,<0.9.0",
        "graphviz<0.9.0,>=0.8.1",
        "mxnet<2.0.0",
        "networkx>=2.3,<3.0",
        "paramiko>=2.4",
        "Pillow>=8.0.1",
        "psutil>=5.7.3",
        "pyarrow>=2.0.0",
        "scikit-optimize",
        "tornado>=6.1",
    ],
)



import numpy as np
import mxnet as mx
from mxnet import gluon
import numbers

from .constants import DATA_TYPE
from .posterior_utils import mxnet_is_ndarray

__all__ = ['ConstantPositiveVector',
           'PositiveScalarEncoding',
           'IdentityScalarEncoding',
           'LogarithmScalarEncoding']


class ConstantPositiveVector(gluon.HybridBlock):
    """
    ConstantPositiveVector
    ======================

    Represents constant vector, with positive entry value represented as Gluon
    parameter, to be used in the context of wrapper classes in
    gluon_blocks.py. Shape, dtype, and context are determined from the
    features argument:

    - If features.shape = (n, d):
       shape = (d, 1) if size_cols = True (number cols of features)
       shape = (n, 1) if size_cols = False (number rows of features)
    - dtype = features.dtype, ctx = features.ctx

    Encoding and internal Gluon parameter:
    The positive scalar parameter is encoded via encoding (see
    ScalarEncodingBase). The internal Gluon parameter (before encoding) has the
    name param_name + '_internal'.
    """

    def __init__(self, param_name, encoding, size_cols, **kwargs):
        super(ConstantPositiveVector, self).__init__(**kwargs)
        assert isinstance(encoding, ScalarEncodingBase)
        self.param_name = param_name
        self.encoding = encoding
        self.size_cols = size_cols
        with self.name_scope():
            init_val_int = encoding.init_val_int
            # Note: The initialization values are bogus!
            self.param_internal = self.params.get(
                param_name + '_internal',
                init=mx.init.Constant(init_val_int),
                shape=(1,), dtype=DATA_TYPE)

    def hybrid_forward(self, F, features, param_internal):
        """Returns constant positive vector

        If features.shape = (n, d), the shape of the vector returned is
        (d, 1) if size_cols = True, (n, 1) otherwise.

        :param F: mx.sym or mx.nd
        :param features: Matrix for shape, dtype, ctx
        :param param_internal: Unwrapped parameter
        :return: Constant positive vector
        """
        # Shape, dtype, ctx is determined by extracting column or row from
        # features, then use ones_like
        axis = 0 if self.size_cols else 1
        ones_vec = F.ones_like(F.reshape(F.slice_axis(
            F.BlockGrad(features), axis=axis, begin=0, end=1), shape=(-1, 1)))
        param = F.reshape(self.encoding.get(F, param_internal), shape=(1, 1))
        return F.broadcast_mul(ones_vec, param)

    def set(self, val):
        self.encoding.set(self.param_internal, val)

    def get(self):
        param_internal = unwrap_parameter(
            mx.nd, self.param_internal, None)
        return self.encoding.get(
            mx.nd, param_internal).asscalar()

    def get_box_constraints_internal(self):
        return self.encoding.box_constraints_internal(
            self.param_internal)

    def log_parameters(self):
        return '{} = {}'.format(self.param_name, self.get())

    def get_parameters(self):
        return {self.param_name: self.get()}

    def switch_updating(self, flag):
        """Is the underlying parameter updated during learning?

        By default, the parameter takes part in learning (its grad_req
        attribute is 'write'). For flag == False, the attribute is
        flipped to 'null', and the parameter remains constant during
        learning.

        :param flag: Update parameter during learning?
        """
        grad_req = 'write' if flag else 'null'
        self.param_internal.grad_req = grad_req

    def has_regularizer(self):
        return (self.encoding.regularizer is not None)

    def eval_regularizer(self, F, features):
        if self.has_regularizer():
            param_internal = unwrap_parameter(
                F, self.param_internal, features)
            param = self.encoding.get(F, param_internal)
            return self.encoding.regularizer(F, param)
        else:
            return 0.0


class ScalarEncodingBase(object):
    """
    ScalarEncodingBase
    ==================

    Base class for encoding and box constraints for Gluon parameter,
    represented as mx.gluon.Parameter. The parameter is with shape (dimension,)
    where dimension is 1 by default.

    An encoding is given as

        param = enc(param_internal), param_internal = dec(param)

    The Gluon parameter represents param_internal, while param is what is
    visible to the outside.

    Here, enc and dec are inverses of each other. enc is used in 'get', dec
    is used in 'set'. Use 'IdentityScalarEncoding' for no encoding (identity).
    NOTE: enc (and dec) must be strictly increasing.

    Box constraints are given by constr_lower_int < constr_upper_int. Here,
    None means no constraint. The constraints apply to param_internal. If both
    are None, param_internal is unconstrained (default).
    NOTE: Box constraints are just maintained here, they have to be enforced
    by an optimizer!

    If regularizer is given, it specifies a regularization term for the
    (encoded) parameter which can be added to a criterion function. It is
    evaluated as regularizer(F, param), where F is mx.sym or mx.nd.

    Typical use cases:
    - Unconstrained optimizer, positive scalar > lower:
      Use PositiveScalarEncoding(lower), box constraints = [None, None]
    - Optimizer supports box constaints [constr_lower, constr_upper]:
      Use IdentityScalarEncoding(constr_lower, constr_upper)
    """

    def __init__(
            self, init_val, constr_lower=None, constr_upper=None,
            regularizer=None, dimension=1):
        if constr_lower is not None and constr_upper is not None:
            assert constr_lower < constr_upper
        init_val = self._check_or_set_init_val(
            init_val, constr_lower, constr_upper)
        init_val_int = self.decode(init_val, 'init_val')
        if constr_lower is not None:
            assert init_val >= constr_lower
            constr_lower_int = self.decode(constr_lower, 'constr_lower')
        else:
            constr_lower_int = None
        if constr_upper is not None:
            assert init_val <= constr_upper
            constr_upper_int = self.decode(constr_upper, 'constr_upper')
        else:
            constr_upper_int = None
        self.constraints = (constr_lower, constr_upper)
        self.constraints_internal = (constr_lower_int, constr_upper_int)
        self.init_val_int = init_val_int
        self.regularizer = regularizer
        self.dimension = dimension

    def get(self, F, param_internal):
        raise NotImplementedError("get must be implemented")

    def set(self, param_internal, param_val):
        assert isinstance(param_internal, mx.gluon.Parameter)
        assert param_internal.shape == (self.dimension,)

        if isinstance(param_val, (list, np.ndarray)):
            assert len(param_val) == self.dimension
            assert np.array(param_val).ndim == 1
            val_int_list = [self.decode(val, 'param_val') for val in param_val]
        else:
            assert np.isscalar(param_val) is True
            val_int_list = [self.decode(param_val, 'param_val')] * self.dimension
        param_internal.set_data(mx.nd.array(val_int_list))

    def decode(self, val, name):
        raise NotImplementedError("decode to be implemented in subclass")

    def box_constraints_internal(self, param_internal):
        assert isinstance(param_internal, mx.gluon.Parameter)
        assert param_internal.shape == (self.dimension,)
        return {param_internal.name: self.constraints_internal}

    def box_constraints(self):
        return self.constraints

    @staticmethod
    def _check_or_set_init_val(init_val, constr_lower, constr_upper):
        if init_val is not None:
            if constr_upper is not None:
                assert init_val <= constr_upper
            if constr_lower is not None:
                assert constr_lower <= init_val
        else:
            # Just pick some value which is feasible
            if constr_lower is not None:
                if constr_upper is not None:
                    init_val = 0.5 * (constr_upper + constr_lower)
                else:
                    init_val = 1.1 * constr_lower
            else:
                if constr_upper is not None:
                    init_val = 0.9 * constr_upper
                else:
                    init_val = 1.0
        return init_val


class PositiveScalarEncoding(ScalarEncodingBase):
    """
    PositiveScalarEncoding
    ======================

    Provides encoding for positive scalar and vector: param > lower.
    Here, param is represented as mx.gluon.Parameter. The param
    is with shape (dimension,) where dimension is 1 by default.

    The encoding is given as:

        param = softrelu(param_internal) + lower,
        softrelu(x) = log(1 + exp(x))

    If constr_upper is used, the constraint

        param_internal < dec(constr_upper)

    can be enforced by an optimizer. Since dec is increasing, this translates
    to param < constr_upper.
    NOTE: While lower is enforced by the encoding, the upper bound is not, has
    to be enforced by an optimizer.
    """

    def __init__(
            self, lower, constr_upper=None, init_val=None, regularizer=None, dimension=1):
        assert isinstance(lower, numbers.Real) and lower >= 0.0
        self.lower = lower
        super(PositiveScalarEncoding, self).__init__(
            init_val, constr_lower=None, constr_upper=constr_upper,
            regularizer=regularizer, dimension=dimension)

    def get(self, F, param_internal):
        return F.Activation(
            param_internal, act_type='softrelu') + self.lower

    def decode(self, val, name):
        assert val > self.lower, \
            '{} = {} must be > self.lower = {}'.format(
                name, val, self.lower)
        # Inverse of encoding: Careful with numerics:
        # val_int = log(exp(arg) - 1) = arg + log(1 - exp(-arg))
        #         = arg + log1p(-exp(-arg))
        arg = val - self.lower
        return arg + np.log1p(-np.exp(-arg))


class IdentityScalarEncoding(ScalarEncodingBase):
    """
    IdentityScalarEncoding
    ======================

    Identity encoding for scalar and vector:

        param = param_internal

    This does not ensure that param is positive! Use this only if positivity
    is otherwise guaranteed.
    """

    def __init__(
            self, constr_lower=None, constr_upper=None, init_val=None,
            regularizer=None, dimension=1):
        super(IdentityScalarEncoding, self).__init__(
            init_val, constr_lower=constr_lower, constr_upper=constr_upper,
            regularizer=regularizer, dimension=dimension)

    def get(self, F, param_internal):
        return param_internal

    def decode(self, val, name):
        return val


class LogarithmScalarEncoding(ScalarEncodingBase):
    """
    LogarithmScalarEncoding
    =======================

    Logarithmic encoding for scalar and vector:

        param = exp(param_internal)
    """

    def __init__(
            self, constr_lower=None, constr_upper=None, init_val=None,
            regularizer=None, dimension=1):
        super(LogarithmScalarEncoding, self).__init__(
            init_val, constr_lower=constr_lower, constr_upper=constr_upper,
            regularizer=regularizer, dimension=dimension)

    def get(self, F, param_internal):
        return F.exp(param_internal)

    def decode(self, val, name):
        assert val > 0.0, \
            '{} = {} must be positive'.format(name, val)
        return np.log(np.maximum(val, 1e-15))


# === Internal ===


def unwrap_parameter(F, param_internal, some_arg=None):
    assert isinstance(param_internal, mx.gluon.Parameter)
    if mxnet_is_ndarray(F):
        ctx = some_arg.context if some_arg is not None else mx.cpu()
        val = param_internal.data(ctx=ctx)
    else:
        val = param_internal.var()
    return val


def encode_unwrap_parameter(F, param_internal, encoding, some_arg=None):
    return encoding.get(F, unwrap_parameter(F, param_internal, some_arg))
